;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil
  (bug-reference-bug-regexp . "#\\(?2:[0-9]+\\)")
  (bug-reference-url-format . "https://github.com/emacs-ess/ess/issues/%s")
  (sentence-end-double-space))
 (emacs-lisp-mode
  (outline-regexp . "\f\\|\\`;\\|;;\\*\\|;;;\\*\\|(def[cvum]\\|(setq\\|;;;;\\*")
  (indent-tabs-mode))
 (ess-r-mode . ((ess-style . RRR))))
