;; Generated package description from ess.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "ess" "25.1.0.0.20260306.163926" "Emacs Speaks Statistics" '((emacs "25.1")) :commit "a8e64e44da01083a74dfdb5bc3e59b444407b1e6" :authors '(("David Smith" . "dsmith@stats.adelaide.edu.au") ("A.J. Rossini" . "blindglobe@gmail.com") ("Richard M. Heiberger" . "rmh@temple.edu") ("Kurt Hornik" . "Kurt.Hornik@R-project.org") ("Martin Maechler" . "maechler@stat.math.ethz.ch") ("Rodney A. Sparapani" . "rsparapa@mcw.edu") ("Stephen Eglen" . "stephen@gnu.org") ("Sebastian P. Luque" . "spluque@gmail.com") ("Henning Redestig" . "henning.red@googlemail.com") ("Vitalie Spinu" . "spinuvit@gmail.com") ("Lionel Henry" . "lionel.hry@gmail.com") ("J. Alexander Branham" . "alex.branham@gmail.com")) :maintainer '("ESS Core Team" . "ESS-core@r-project.org") :url "https://ess.r-project.org/")
